// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
// 
// Module: BinnedProxyResponse
// Author: Hassan Siddiqui
//
// Description: 
// Provides a proxy layer to any GenericResponse object.
// This later rebins the raw data to suit the rebin
// factor specified.
// 
// Comments: 
//  
// Modifications: See ChangeLog
// 
// $Id: BinnedProxyResponse.cc,v 1.17 2007/06/06 12:42:14 eojero Exp $

/* General */

/* SAS-specific */
#include "Tracer.h"

/* Local */
#include "BinnedProxyResponse.h"
#include "IntervalRebinner.h"
#include "VectorRebinner.h"
#include "NonUnifBinnedAxis.h"

/********************************
 *
 * class BinnedProxyResponse
 *
 * Provides a proxy layer to any GenericResponse object.
 * This later rebins the raw data to suit the rebin
 * factor specified.
 *
 *********************************/

/* Constructors */
BinnedProxyResponse::BinnedProxyResponse (GenericResponse * response,
					  unsigned binfactor,
					  unsigned offset,
					  unsigned firstchannel,
					  unsigned detchans):
  _response(response),
  _binFactor(binfactor),
  _offset(offset),
  _firstChannel(firstchannel),
  _detchans(detchans)
{
  PROC;
  
  _lastChannel  = _firstChannel + _detchans - 1;

  /* Set ebounds */
  _ebounds = 
    new NonUnifBinnedAxis(IntervalRebinner::rebin(_response->ebounds(),
						  _binFactor,
						  _offset));
  ENDPROC;
}

/* Destructors */
BinnedProxyResponse::~BinnedProxyResponse()
{
  PROC;
  delete _ebounds;
  ENDPROC;
}


/** Channel range accessors */
const unsigned BinnedProxyResponse::firstChannel() const
{
  PROC;
  RETURN _firstChannel;
}

const unsigned  BinnedProxyResponse::lastChannel() const
{
  PROC;
  RETURN _lastChannel;
}


/** Response Accessors */
unsigned BinnedProxyResponse::ebinFirstIndex() const
{
  PROC;
  RETURN _response->ebinFirstIndex();
}

unsigned BinnedProxyResponse::numberOfEbins() const
{
  PROC;
  RETURN _response->numberOfEbins();
}

const Bin BinnedProxyResponse::ebins(unsigned int row) const
{
  PROC;
  RETURN _response->ebins(row);
}

const vector<double> BinnedProxyResponse::responseVector ( unsigned int row ) const
{
  PROC;
  errstr << "BinnedProxyResponse::responseVector ( unsigned int ) called for row " << row
	 << Sas::message(Msg::App, Msg::Verbose);

  vector<double> out = 
    VectorRebinner::rebin(_response->responseVector(row),
			  _binFactor,
			  _offset);

  if (out.size() != _detchans)
    {
      errstr << "Number of channels in rebinned response ["
	     << out.size()
	     << "] differs to that expected ["<<_detchans<<"]"
	     << Sas::error("BinnedProxyReponseInvalidResponse");
    }

  errstr << "BinnedProxyResponse::responseVector ( unsigned int ) complete. "
	 << Sas::message(Msg::App, Msg::Verbose);
  
  RETURN out;
}

const vector<double> BinnedProxyResponse::responseVector ( double energy ) const
{
  PROC;

  errstr << "BinnedProxyResponse::responseVector ( double energy ) called.. "
	 << Sas::message(Msg::App, Msg::Verbose);
  
  
  vector<double> out = 
    VectorRebinner::rebin(_response->responseVector(energy), 
			  _binFactor,
			  _offset);

  if (out.size() != _detchans)
    {
      errstr << "Number of channels in rebinned response ["
	     << out.size()
	     << "] differs to that expected ["<<_detchans<<"]"
	     << Sas::error("BinnedProxyReponseInvalidResponse");
    }
  
  errstr << "BinnedProxyResponse::responseVector ( double energy ) complete. "
	 << Sas::message(Msg::App, Msg::Verbose);
  
  RETURN out;
}

/* Ebounds Accessor */
const BinnedAxis * BinnedProxyResponse::ebounds() const
{
  PROC;
  RETURN _ebounds;
}

unsigned int BinnedProxyResponse::energyToPha( double energy ) const
{
  PROC;


  unsigned channel_in  = _response->energyToPha(energy);

  if (channel_in < _offset)
    {
      errstr << "BinnedProxyResponse::energyToPha: channel_in ["
	     << channel_in
	     << " is less than offset " << _offset
	     << Sas::error("BinnedProxyResponseInvalidChannel");
    }

  unsigned temp = channel_in - _offset;
  unsigned channel_out = (temp)/_binFactor;
  
  channel_out = (channel_out*_binFactor == temp) ?
    channel_out : channel_out + 1;

  RETURN channel_out;
}
