// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
// 
// Module: CorePosData
// Author: Richard Saxton
//
// Description: Contains all the code to extract and manipulate
//              positions and attitude information
// 
// Comments: 
//  
// Modifications: See ChangeLog
// 
/* General */

#include "Xmm.h"
#include "common.h"
#include "OalServer.h"
#include "CalServer.h"
#include "DataSubspace.h"
#include "DssBuilder.h"
#include "Tracer.h"
#include "CcdQuantumDataServer.h"
#include "ExceptionErrorHandler.h"
#include "CheckExpr.h"
#include "SelectUtils.h"

/* Local classes */
#include "CorePosData.h"
#include "RegionExtent.h"
#include "PixCoords.h"

/* Constructors */

CorePosData::CorePosData( DataSet * spectrumset , Table * spectrumtable, 
    DatafileAttributes * specAttributes) 
{
  PROC; 

  /* Read info from the header of the spectrum extension */
  _dss = new DataSubspace(spectrumtable);

  _instrumentID =  specAttributes->instrumentID();

  /* Set up a coordinate conversion object */
  _poswcs = new PosWCS(spectrumset, specAttributes->rapnt(), 
                  specAttributes->decpnt(), specAttributes->papnt(), false);

  /* Create a position utilities object */
  _positionUtils = new PositionUtils(_instrumentID);

  

    errstr
    << "Set up poswcs and positionUtils"
    << Sas::message(Msg::App, Msg::Verbose);


  /* For timing and burst modes the spatial selection is either left blank
     or is made in RAW-X. In this case a pseudo-region should be generated
     in DET coords to handle vignetting and PSF issues */
  Xmm::Mode windowMode = specAttributes->getWindowMode(); 

  if (windowMode == Xmm::FAST_TIMING || windowMode == Xmm::FAST_BURST
                                     || windowMode == Xmm::FAST_UNCOMPRESSED
                                     || windowMode == Xmm::FAST_COMPRESSED) {

     addDetRegion(_positionUtils);
  }

  /* If the spatial region consists of a single mask add a large box to the
     DSS in either sky or detector coordinates */
  bool maskFound=false;
  DSS::DSaxes axesdet,axespos;
  axesdet.push_back("DET");
  axespos.push_back("POS");
                                                                                
  if ( CheckExpr::hasMask(_dss->selectExpression(axespos)) ||
       CheckExpr::hasMask(_dss->selectExpression(axesdet)) ) {
                                                                                
    maskFound=true;
    std::string expression=" ";
                                                                                
    if ( CheckExpr::hasMask(_dss->selectExpression(axespos))) {
      expression = "(X,Y) in BOX(25921,25921,25920,25920,0)";
    } else {
      expression = "(DETX,DETY) in BOX(0,0,25920,25920,0)";
    }
                                                                                
    DataSubspace _dssextra = dssBuilder.getDss(expression);
    for(unsigned icomp=0; icomp <_dssextra.components(); icomp++){
      _dss->addComponent(*_dssextra.component(icomp));
    }
                                                                                
    errstr << "New total Expression "<<_dss->selectExpression(axespos)
         << Sas::message(Msg::App, Msg::Noisy);
  }
                                                                                
  /* Create a source box extent object */
  RegionExtent * extent = new RegionExtent(_dss, _positionUtils, 
                               specAttributes->datamode(), 
                               specAttributes->observationDate(), _poswcs);

  /* Get the source box centre detector coordinates */
  _sourcedetx = (extent->xmin() + extent->xmax()) / 2.0;
  _sourcedety = (extent->ymin() + extent->ymax()) / 2.0;

     errstr 
       << "source position: "<<_sourcedetx<<" "<<_sourcedety
       << Sas::message(Msg::App, Msg::Verbose);

  /* Get the time ranges used to create this spectrum */
  _timeList = setTimeList(spectrumset);


ENDPROC;
};


/*  Destructors  */
CorePosData::~CorePosData()
{
  PROC;

  delete _positionUtils;   _positionUtils  = 0;
  delete _poswcs;   _poswcs = 0;
  delete _dss;   _dss = 0;

  ENDPROC;
};

/* Return the source centre in tel coords (theta,phi) */
CoordPair CorePosData::centre_detcoord() const
{
  return CoordPair(_sourcedetx,_sourcedety);
}
  
/* Return the source centre in tel coords (theta,phi) */
TelCoords CorePosData::centre_telcoord() const
{
  return _positionUtils->camCoord2ToTelCoord(
              _positionUtils->detxy005ToCamCoord2(_sourcedetx,_sourcedety));
}
  
/* Return the source centre in raw coordinates */
PixCoords CorePosData::centre_raw() const
{
  
  PixCoords pixCoord(0,0);

  errstr 
       << "source position: "<<_sourcedetx<<" "<<_sourcedety
       << Sas::message(Msg::App, Msg::Verbose);

  /* Check if the source centre lies on a chip */
  CamCoords testpixel = _positionUtils->detxy005ToCamCoord2(
                                           _sourcedetx,_sourcedety);

  GeometryDataServer * ct = 0; ct = calServer.getAtom(ct);

  /* Find extents of CCD depending on instrument */
  unsigned minx=1, maxx=600, miny=1, maxy=600; // MOS values

  if (_instrumentID == Xmm::EPN)
     {
      minx=1; maxx=199; miny=1; maxy=199;
     }

  /* Return the RAW X/Y coords of the region centre */
  if ( ct->onCcd(testpixel.x(), testpixel.y()) == Xmm::CCD_NONE ) 
  {

     if (_instrumentID == Xmm::EPN) {

      /* Source is off-chip so set the raw coordinates as 32.0, the Y
       value calculated from the DET-Y position */

       double rawX = 32.0;
       double rawY = 199.0 - abs(_sourcedety + 1000) * 199.0 / 16000.0;
       rawY = rawY < 1 ? 1:rawY;
       rawY = rawY > 199 ? 199:rawY;

       pixCoord = PixCoords(UNT(rawX), UNT(rawY));

       errstr 
         << "source is not centred on a CCD - generating RMF for "
         << "line number "<< rawY
         << Sas::message(Msg::App, Msg::Verbose);
     }
     else // MOS
     {
       pixCoord = PixCoords(minx, miny);
     }
  }
  else 
  {
     PixCoords testpixCoord = _positionUtils->camCoord2ToPixCoord(testpixel);
     unsigned xpix = testpixCoord.x();
     unsigned ypix = testpixCoord.y();
     xpix = xpix < minx ? minx:xpix;
     xpix = xpix > maxx ? maxx:xpix;
     ypix = ypix < miny ? miny:ypix;
     ypix = ypix > maxy ? maxy:ypix;
     pixCoord = PixCoords(xpix,ypix);
  }

  RETURN  pixCoord;
}

/* Return the CCD number for the source centre */
int CorePosData::centre_CCD() const
{

  /* Convert the centre position into CamCoords */
  CamCoords testpixel = _positionUtils->detxy005ToCamCoord2(
                                           _sourcedetx,_sourcedety);

  /* USe the geometry server to find the correct CCD */
  GeometryDataServer * ct = 0; ct = calServer.getAtom(ct);

  int ccd = ct->onCcd(testpixel.x(), testpixel.y());

  /* Set the CAL state to this pixel */
  calServer.state()->ccd()->set(ccd);

  errstr<<"Test pixel centred on CCD: "<<ccd
        << Sas::message(Msg::App, Msg::Verbose);

  /* Return the CCD number from the CAL */
  return ccd;
}

/* Get selected patterns from DSS */
EventGrades CorePosData::patternCode() const
{
  PROC;

  EventGrades pcode=CombinedEventGrades::SDTQ;
  double lowest=12, highest=0;
  double upper, lower;
  RangeBound upperB, lowerB;

  list <Range<double> > ranges = _dss->ranges("PATTERN");

  if (ranges.size()==0)
  {
      errstr 
	<< "There is no PATTERN range info in the DSS. Assume the full range of patterns has been used."
	<< Sas::warning("NoDSSPatterns");

     /* Default to singles, doubles, triples and quads for MOS and
        singles + doubles for PN*/
      if (_instrumentID == Xmm::EPN)
      {
        pcode=CombinedEventGrades::SD;
      }
      else {
        pcode=CombinedEventGrades::SDTQ;
      }
  
  } else {

   /* Find the lowest and highest pattern values wanted */
    for ( list<Range<double> >::iterator i = ranges.begin(); 
	i != ranges.end(); 
	++i )
     {
      (*i).values(lower, upper, lowerB, upperB);

      if (lowerB==Unbounded)
	{
	  lower = 0;
	}
      else if (lowerB==Open)
        {
          ++lower;
        }

      if (upperB==Unbounded)
	{
	  upper = 12;
	}
      else if (upperB==Open)
        {
          --upper;
        }
	  
      lowest = lower < lowest ? lower:lowest;
      highest = upper > highest ? upper:highest;
     }
  
     errstr 
       << "Pattern range in file: "<<lowest<<":"<<highest
       << Sas::message(Msg::App, Msg::Verbose);

     if (lowest==0 && highest==0)
       {
        //pcode=PatternFractionDataServer::SINGLE;
        pcode=CombinedEventGrades::SINGLE;
       }
     else if (lowest==0 && highest==4)
       {
        pcode=CombinedEventGrades::SD;
       }
     else if (lowest==0 && highest==12)
       {
        pcode=CombinedEventGrades::SDTQ;
       }
     else if (lowest==1 && highest==4)
       {
        pcode=CombinedEventGrades::DOUBLE;
       }
     else if (lowest==5 && highest==8)
       {
        pcode=CombinedEventGrades::TRIPLE;
       }
     else if (lowest==9 && highest==12)
       {
        pcode=CombinedEventGrades::QUADRUPLE;
       }
     else
       {
          errstr
        << "An RMF cannot be produced for the pattern range in the DSS"
        << Sas::error("InvalidPatterns");
       }
   }

  /* need to put a temporary fiddle in here for MOS Timing mode.
     MOS Timing mode spectra are produced from pattern 0 events but
     these actually combine image mode pattern 0,1 and 3 events.
     This should be handled within the EMOSn_REDIST CCF file but
     currently isn't so here we will fix the response to be Singles+dubs */

  /* REMOVE ?????
   if ( (_instrumentID == Xmm::EMOS1 || _instrumentID == Xmm::EMOS2) && 
          calServer.state()->mode()->get() == Xmm::FAST_UNCOMPRESSED) {
      pcode=CombinedEventGrades::SD;
   }
  */

   // Throw an error if MOS spectrum with anything other than S or SDTQ as fails later
   if ( (_instrumentID == Xmm::EMOS1 || _instrumentID == Xmm::EMOS2) &&
       ! (pcode==CombinedEventGrades::SINGLE || pcode==CombinedEventGrades::SDTQ) )
          errstr<<"An RMF cannot be produced for MOS with the pattern range in the DSS"
                << Sas::error("InvalidPatterns");

   // DEBUG
   errstr << "Pattern code="<<pcode
          << Sas::message(Msg::App, Msg::Verbose);
       

   RETURN pcode;

  ENDPROC;
}

/* addDetRegion - adds a pseudo spatial region in detector coords to the dss */
void CorePosData::addDetRegion(PositionUtils * putils)
{
  PROC;

  /* For timing and burst modes the detector position is unimportant
     so dummy up a position in detector coords in the centre of the
     chip */

    double lower_rawx=30, lower_rawy=100;
    double upper_rawx=50, upper_rawy=120;

   /* Now find the detector coords of this position */
    PixCoords pixCoord1(UNT(lower_rawx), UNT(lower_rawy));
    CamCoords camCoord2 = putils->PixCoordTocamCoord2(pixCoord1);
    CoordPair detlow = putils->CamCoord2Todetxy005(camCoord2);
    double det_xlow = detlow.x();
    double det_ylow = detlow.y();
    PixCoords pixCoord2(UNT(upper_rawx), UNT(upper_rawy));
    camCoord2 = putils->PixCoordTocamCoord2(pixCoord2);
    CoordPair detupp = putils->CamCoord2Todetxy005(camCoord2);
    double det_xupp = detupp.x();
    double det_yupp = detupp.y();

    double detxcen = (det_xlow + det_xupp) / 2.0;
    double detycen = (det_ylow + det_yupp) / 2.0;
    double detx_hwid = abs(detxcen - det_xlow);
    double dety_hwid = abs(detycen - det_ylow);

    errstr<<"Det pixel centre and width "<<detxcen<<" "<<detycen
              <<detx_hwid<<" "<<dety_hwid
              << Sas::message(Msg::App, Msg::Noisy);

   /* Create an expression from these coords */
    std::string expression;
   // expression="(DETX,DETY) in BOX("+toString(detxcen)+","+toString(detycen)+
    //           ","+toString(detx_hwid)+","+toString(dety_hwid);
    expression="(DETX,DETY) in BOX(100,100,100,100";
    expression += ",0)";

    errstr<<"From raw X: "<<lower_rawx<<" "<<upper_rawx
              <<" raw Y: "<<lower_rawy<<" "<<upper_rawy<<" got: "
              <<expression
              << Sas::message(Msg::App, Msg::Verbose);

   /* Create a filter */
    DataSubspace _dssextra = dssBuilder.getDss(expression);

    for(unsigned icomp=0; icomp <_dssextra.components(); icomp++){

      errstr<<"Adding component "<<icomp<<" to DSS"
              << Sas::message(Msg::App, Msg::Verbose);

      _dss->addComponent(*_dssextra.component(icomp));
    }

    errstr<<"Added new component to DSS"
              << Sas::message(Msg::App, Msg::Verbose);

 ENDPROC;
}

/*
* Method: setTimeList(Dataset)
* Purpose: Find the ranges of times used within the DSS
*/
  GtiList CorePosData::setTimeList(DataSet * spectrumset) const
  {

    GtiList mergedList, newlist;

   // Find the STDGTIs for this CCD 
    int ccd = centre_CCD();

   // Loop over each DSS component relating to a GTI for the central CCD
    unsigned ncomps = _dss->components();

    for(unsigned icomp=0; icomp<ncomps; icomp++)
      {
        DScomp * comp = _dss->component(icomp);
        for(unsigned ifilt=0; ifilt<comp->filters(); ifilt++){
            DSfilter * filter = comp->filter(ifilt);

           //GTI filter
            if (filter->type()==DSS::GtiT)
            {
              // Find the CCD number which this filter applies to
              // taken from epiclccorr code (searchFor)
               std::vector<std::string> vec(breakup(filter->blockSpec(), ":"));
               std::stringstream ss;
               int ccd_number;

              // Check if we have a STDGTI block or a GTI block
               if (vec[1].find("STDGTI") != string::npos)
                {
                  ss << vec[1][vec[1].size()-2] << vec[1][vec[1].size()-1];
                  ss >> ccd_number;
                }
               else
                {
                  Block* expBlock = spectrumset->block(vec[1]);

                  if (!expBlock->hasAttribute("CCDID"))
                   {
                     errstr<<"No CCDID keyword found in table "
                           <<vec[1]<< Sas::error("NoCCDID");
                   }
                  else
                   {
                     ccd_number = expBlock->attribute("CCDID")->asInt();
                   }
                   expBlock->release();
                }

             
                if (ccd_number == ccd)
                {
                   errstr<<"Processing GTI extension: "<<vec[1]
                         <<" corresponding to CCD " <<ccd_number
                         << Sas::message(Msg::App, Msg::Verbose);

                   string tabname = spectrumset->name()+":"+vec[1];
                   InOGIPgtiTable in1 = InOGIPgtiTable(tabname);

                  // Extract these GIs into a list and merge all together
                   in1>>newlist;
                   mergedList |= newlist;
                }

            }
        }
    }
    mergedList.sortIt();

    return mergedList;
}

/*
* Method: timeList()
* Purpose: Return the ranges of times used in the creation of the spectrum
*/
  GtiList CorePosData::getTimeList() const
{
   return _timeList;
}
