// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
// This  skeleton file was automatically created by 
//            pkgmaker 0.1
//            Thursday, December 11 1997
// 
// Module: PsfImage
// Author: Richard Saxton
//
// Description: 
// Creates a Psf image for each pattern weighted by the input energy spectrum
// 
// Comments: 
//  
// Modifications: See ChangeLog
// 
// $Id: PsfImage.cc,v 1.3 2013/03/12 14:14:09 rsaxton Exp $

// General classes
#include<cassert>

#include <string>
#include <iostream>
#include <vector>
#include <map>

// SAS-specific classes

#include "Dal.h"
#include "errstr.h"
#include "EnergySpectrum.h"
#include "PsfImage.h"
#include "CalOalUtility.h"
#include "PattFracs.h"
#include "PixelIterator.h"

// Utility functions
#include "Tracer.h"

// -----------------------------------------------------------------------------------

// class PsfImage

// Supports data consisting of an energy, energy width and data value

// -----------------------------------------------------------------------------------

 /* Constructor - from a fluxed spectrum */
PsfImage::PsfImage(EnergySpectrum input_spectrum, CoordPair halfWidthmm,
                   PattFracs *pfrac, double theta, double phi)
{
  PROC;

    init(input_spectrum, halfWidthmm, pfrac, theta, phi);

}

 /* Constructor - from a single energy */
PsfImage::PsfImage(double energy, PattFracs *pfrac, 
                                     double theta, double phi)
{
  PROC;

  /* Make an input energy spectrum for one energy */
    vector<float> energyVec(1,float(energy));
    vector<float> energy_width(1,1.0); 
    vector<float> strength(1,1.0);  
    EnergySpectrum e = EnergySpectrum(energyVec, energy_width, strength);

  /* Find the maximum PSF extent from the server - but now default to 5' */
    CoordPair halfWidthmm = CoordPair( 300.0 * 0.036367,
                            300.0 * 0.036367);
  
    init(e, halfWidthmm, pfrac, theta, phi);
}

 /* Initialisation */
void PsfImage::init(EnergySpectrum input_spectrum, CoordPair halfWidthmm,
                   PattFracs *pfrac, double theta, double phi)
{

  /* Access the PSF data server */
    PsfDataServer *pds = 0; pds = calServer.getAtom(pds);

  /* Get the pattern fraction as a function of energy at this chip position */

  /* Create a window to contain the whole source region */
    SpatialWindow<double> window = SpatialWindow<double>(halfWidthmm.x(),
                                                         halfWidthmm.y());

  /* Get the PSFs as an energy-indexed array */
    vector<Image<float> > psfimage_energy;

    for (unsigned e=0; e<input_spectrum.energy().size(); ++e) 
     {
        Psf &psf = pds->psf(input_spectrum.energy()[e], theta, phi); 

        psfimage_energy.push_back(psf.image(window));
      // Clean up psf ??
     }

    _nx = psfimage_energy[0].xExtent(), 
    _ny = psfimage_energy[0].yExtent();
    
  /* Calculate a pattern-specific PSF which weights for the energy spectrum
     and for the pattern fractions at each energy */
    Image<float> psfimage_temp(_nx,_ny); // For singles, dubs, t and q

  // Loop over each pattern type - singles, doubles, trips and quads
    for (unsigned p=0; p<4; ++p)
     {
     // Set the image for this pattern to 0
      for (unsigned x=0; x<_nx; ++x)
       for (unsigned y=0; y<_ny; ++y)
          psfimage_temp[x][y]=0.0;

     // Loop over each energy in the input energy spectrum 
      for (unsigned e=0; e<input_spectrum.energy().size(); ++e) 
       {
        // Calculate the pattern fraction for this pattern and energy
        // Also calculate the spectral weighting
         double f = pfrac->fraction(p,input_spectrum.energy()[e]); 
         double w = input_spectrum.value(e);

        // Add the PSF image at this energy weighted by the pattern 
        // fraction for this pattern at this energy weighted by the spectrum
        // value at this energy
         for (unsigned x=0; x<_nx; ++x)
          for (unsigned y=0; y<_ny; ++y)
             psfimage_temp[x][y] += psfimage_energy[e][x][y] * f * w;
       }
       _psfimage_pattern.push_back(normalise(psfimage_temp));
     }

  // Clean up
    calServer.releaseAtom(pds);
}

/* Destructor */
PsfImage:: ~PsfImage()
{
  PROC;
  ENDPROC;
};

// Method: image - returns the PSF image for the requested pattern
Image<float> & PsfImage::image(int pattern) 
{

     assert(pattern<4);

     return _psfimage_pattern[pattern];
}

// Method: normalise - returns an input image normalised to one
Image<float> PsfImage::normalise(Image<float> &im) 
{

   double finaltot=0.0;
   for (unsigned ix=0; ix<_nx; ++ix)
     for (unsigned iy=0; iy<_ny; ++iy) 
          finaltot += im[ix][iy];
            
   if (finaltot > 0.0) {
     for (unsigned ix=0; ix<_nx; ++ix)
       for (unsigned iy=0; iy<_ny; ++iy) 
            im[ix][iy] = im[ix][iy] / finaltot;
   }

   return im;
}

// Method: normalise() - normalises all the internal pattern images
void PsfImage::normalise()
{
    for (unsigned p=0; p<4; ++p)
       normalise(_psfimage_pattern[p]);
}

// Method: cleanRegion - sets pixels outside a region to zero
// Inputs: pxitr - an iterator for pixels in the source region
//         detoff - the centre of the source region in detector coords
void PsfImage::cleanRegion(PixelIterator * pxitr, CoordPair detoff)
{

  // Coordinate declarations
    int xhalf = INT (_nx / 2) ;
    int yhalf = INT (_ny / 2);
    int xmax = _nx-1, ymax = _ny-1;
    double binsize = 20.0;

  // Declare a temporary image
    Image<float> psfimage_temp(_nx,_ny); 

  // Set the temporary image for this pattern to 0
    for (unsigned x=0; x<_nx; ++x)
      for (unsigned y=0; y<_ny; ++y)
        psfimage_temp[x][y]=0.0;

  // Loop over each pixel in the input region 
    FOREACH(pxitr)
      {
        PixelElement pixel = pxitr->currentItem();

       // Convert from input detector coords into PSF-centred arcsecs frame
        int xlo = round((pixel.xCentre()-detoff.x()-pixel.xWidth()/2.0)  
                                  / binsize) + xhalf;
        int xhi = round((pixel.xCentre()-detoff.x()+pixel.xWidth()/2.0) 
                                  / binsize) + xhalf;
        int ylo = round((pixel.yCentre()-detoff.y()-pixel.yWidth()/2.0)
                                  / binsize) + yhalf;
        int yhi = round((pixel.yCentre()-detoff.y()+pixel.yWidth()/2.0)
                                  / binsize) + yhalf;

       // Set the temp image pixel to 1 corresponding to PSF pixels
       // which lie within the region pixel 
        for (int x=xlo; x<=xhi; ++x)
          for (int y=ylo; y<=yhi; ++y)
            if (x >= 0 && x <=xmax && y >= 0 && y <=ymax) 
               { psfimage_temp[x][y]=1.0; }       

 //     cout<<"Cam Coord: "<<pixel.xCentre()<<" "<<pixel.xWidth()<<" "<<detoff.x()<<" "<<binsize<<" "<<xhalf<<" "<<_nx<<" "<<_ny<<" "<<xlo<<" "<<xhi<<" "<<ylo<<" "<<yhi<<" "<<endl;

      }

  // Loop over each pattern type - singles, doubles, trips and quads
    for (unsigned p=0; p<4; ++p)
     {
       // Set pixels outside the region to 0 by multiplying by temp image
        for (unsigned x=0; x<_nx; ++x)
         for (unsigned y=0; y<_ny; ++y)
            _psfimage_pattern[p][x][y] *= psfimage_temp[x][y];
     }
}
